package ru.dbotthepony.mc.otm.client.render

import com.google.gson.JsonDeserializationContext
import com.google.gson.JsonDeserializer
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSerializationContext
import com.google.gson.JsonSerializer
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.core.linearInterpolation
import java.lang.reflect.Type

sealed class AbstractSkinElement {
	abstract val x: Float
	abstract val y: Float
	abstract val width: Float
	abstract val height: Float
	abstract val u0: Float
	abstract val v0: Float
	abstract val u1: Float
	abstract val v1: Float
	open val winding: UVWindingOrder get() = UVWindingOrder.NORMAL

	abstract val texture: ResourceLocation

	/**
	 * See [ru.dbotthepony.mc.otm.client.render.clearDepth]
	 */
	fun clearDepth(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
	) = ru.dbotthepony.mc.otm.client.render.clearDepth(stack, x, y, width, height)

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRaw(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = render(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRawPartial(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderWidth(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		winding: UVWindingOrder = this.winding
	) {
		render(stack, x, y, width = width, winding = winding)
	}

	@JvmOverloads
	fun renderWidth(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderWidth(stack, x.toFloat(), y.toFloat(), width.toFloat(), winding)

	@JvmOverloads
	fun renderHeight(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		render(stack, x, y, height = height, winding = winding)
	}

	@JvmOverloads
	fun renderHeight(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderHeight(stack, x.toFloat(), y.toFloat(), height.toFloat(), winding)


	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderRaw(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		val u1 = (u0 + linearInterpolation(width / this.width, 0f, u1 - u0)).coerceAtLeast(0f).coerceAtMost(1f)
		val v1 = (v0 + linearInterpolation(height / this.height, 0f, v1 - v0)).coerceAtLeast(0f).coerceAtMost(1f)

		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderRawPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	abstract fun toJson(): JsonElement
	abstract fun toNetwork(buff: FriendlyByteBuf)

	companion object : TypeAdapter<AbstractSkinElement>(), JsonSerializer<AbstractSkinElement>, JsonDeserializer<AbstractSkinElement> {
		fun toNetwork(value: AbstractSkinElement, buff: FriendlyByteBuf) {
			buff.writeBoolean(value is AtlasSkinElement)
			value.toNetwork(buff)
		}

		fun fromNetwork(buff: FriendlyByteBuf): AbstractSkinElement {
			if (buff.readBoolean()) {
				return AtlasSkinElement.fromNetwork(buff)
			}

			return SkinElement.fromNetwork(buff)
		}

		fun fromJson(json: JsonElement): AbstractSkinElement {
			if (json is JsonPrimitive) {
				return AtlasSkinElement.fromJson(json)
			} else if (json is JsonObject) {
				return SkinElement.fromJson(json)
			} else {
				throw JsonSyntaxException("Unknown element type ${json::class.qualifiedName}")
			}
		}

		override fun write(out: JsonWriter, value: AbstractSkinElement) {
			TypeAdapters.JSON_ELEMENT.write(out, value.toJson())
		}

		override fun read(`in`: JsonReader): AbstractSkinElement {
			return fromJson(TypeAdapters.JSON_ELEMENT.read(`in`))
		}

		override fun serialize(
			src: AbstractSkinElement,
			typeOfSrc: Type,
			context: JsonSerializationContext
		): JsonElement {
			return src.toJson()
		}

		override fun deserialize(
			json: JsonElement,
			typeOfT: Type,
			context: JsonDeserializationContext
		): AbstractSkinElement {
			return fromJson(json)
		}
	}
}
