package ru.dbotthepony.mc.otm.client.render

import com.google.gson.JsonDeserializationContext
import com.google.gson.JsonDeserializer
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSerializationContext
import com.google.gson.JsonSerializer
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.core.set
import java.lang.reflect.Type

fun ResourceLocation.element(
	x: Float,
	y: Float,
	width: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, height, textureWidth, textureHeight)

fun ResourceLocation.pixel(
	x: Float,
	y: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, 1f, textureWidth, textureHeight)

fun ResourceLocation.hLine(
	x: Float,
	y: Float,
	width: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, 1f, textureWidth, textureHeight)

fun ResourceLocation.vLine(
	x: Float,
	y: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, height, textureWidth, textureHeight)

@Suppress("unused")
data class SkinElement @JvmOverloads constructor(
	override val texture: ResourceLocation,
	override val x: Float,
	override val y: Float,
	override val width: Float,
	override val height: Float,
	val imageWidth: Float = 256f,
	val imageHeight: Float = 256f,
	override val winding: UVWindingOrder = UVWindingOrder.NORMAL,
) : AbstractSkinElement() {
	override fun toJson(): JsonObject {
		return JsonObject().also {
			it["texture"] = JsonPrimitive(texture.toString())
			it["x"] = JsonPrimitive(x)
			it["y"] = JsonPrimitive(y)
			it["width"] = JsonPrimitive(width)
			it["height"] = JsonPrimitive(height)
			it["imageWidth"] = JsonPrimitive(imageWidth)
			it["imageHeight"] = JsonPrimitive(imageHeight)
			it["winding"] = JsonPrimitive(winding.name)
		}
	}

	override fun toNetwork(buff: FriendlyByteBuf) {
		buff.writeUtf(texture.toString())
		buff.writeFloat(x)
		buff.writeFloat(y)
		buff.writeFloat(width)
		buff.writeFloat(height)
		buff.writeFloat(imageWidth)
		buff.writeFloat(imageHeight)
		buff.writeEnum(winding)
	}

	companion object : TypeAdapter<SkinElement>(), JsonSerializer<SkinElement>, JsonDeserializer<SkinElement> {
		fun fromNetwork(buff: FriendlyByteBuf): SkinElement {
			val texture = ResourceLocation(buff.readUtf())
			val x = buff.readFloat()
			val y = buff.readFloat()
			val width = buff.readFloat()
			val height = buff.readFloat()
			val imageWidth = buff.readFloat()
			val imageHeight = buff.readFloat()
			val winding = buff.readEnum(UVWindingOrder::class.java)

			return SkinElement(texture, x, y, width, height, imageWidth, imageHeight, winding)
		}

		fun fromJson(value: JsonObject): SkinElement {
			val texture = value["texture"]?.asString ?: throw JsonSyntaxException("Missing texture element")
			val x = value["x"]?.asFloat ?: throw JsonSyntaxException("Missing x element")
			val y = value["y"]?.asFloat ?: throw JsonSyntaxException("Missing y element")
			val width = value["width"]?.asFloat ?: throw JsonSyntaxException("Missing width element")
			val height = value["height"]?.asFloat ?: throw JsonSyntaxException("Missing height element")
			val imageWidth = value["imageWidth"]?.asFloat ?: throw JsonSyntaxException("Missing imageWidth element")
			val imageHeight = value["imageHeight"]?.asFloat ?: throw JsonSyntaxException("Missing imageHeight element")
			val winding = value["winding"]?.asString ?: throw JsonSyntaxException("Missing winding element")

			return SkinElement(ResourceLocation(texture), x, y, width, height, imageWidth, imageHeight, UVWindingOrder.valueOf(winding))
		}

		fun fromJson(value: JsonElement): SkinElement {
			return fromJson(value as? JsonObject ?: throw JsonSyntaxException("Expected JsonObject, got ${value::class.qualifiedName}"))
		}

		override fun write(out: JsonWriter, value: SkinElement) {
			TypeAdapters.JSON_ELEMENT.write(out, value.toJson())
		}

		override fun read(reader: JsonReader): SkinElement {
			val read = TypeAdapters.JSON_ELEMENT.read(reader)
			return fromJson(read as? JsonObject ?: throw JsonSyntaxException("Expected JsonObject, got ${read::class.qualifiedName}"))
		}

		override fun serialize(src: SkinElement, typeOfSrc: Type, context: JsonSerializationContext): JsonElement {
			return src.toJson()
		}

		override fun deserialize(
			json: JsonElement,
			typeOfT: Type,
			context: JsonDeserializationContext
		): SkinElement {
			return fromJson(json as? JsonObject ?: throw JsonSyntaxException("Expected JsonObject, got ${json::class.qualifiedName}"))
		}
	}

	init {
		require(x >= 0f) { "Invalid x $x" }
		require(y >= 0f) { "Invalid y $y" }
		require(width > 0f) { "Invalid width $width" }
		require(height > 0f) { "Invalid height $height" }
		require(imageWidth > 0f) { "Invalid image width $imageWidth" }
		require(imageHeight > 0f) { "Invalid image height $imageHeight" }
	}

	override val u0 = this.x / imageWidth
	override val v0 = this.y / imageHeight
	override val u1 = (this.x + width) / imageWidth
	override val v1 = (this.y + height) / imageHeight
}
